import httpclicent from '@/apis/request'
import pluralize from 'pluralize';
import { getModelValue } from './model';

class CRUD {
    constructor(opts) {
        opts = Object.assign({
            moduleName: '',
            tableName: '',
            apiPrefix: '',
            schemas: [],
        }, opts);
        this.opts = opts
        this.primaryKey = ''
        this.modelName = opts.moduleName
        this.tableName = opts.tableName
        this.apiPrefix = opts.apiPrefix || opts.moduleName || 'rest'
        this.models = []
        this.sortable = []
        this.queryParams = {}
        this.fixedQueueParams = {}
        this.pagination = {
            index: 1,
            size: 15,
            totalCount: 0,
        }
        //初始化schema
        if (Array.isArray(opts.schemas)) {
            this.schemas = opts.schemas
        } else if (!Array.isArray(opts.schemas) && typeof opts.Array === 'object') {
            this.schemas = [];
            for (let k in opts.schemas) {
                this.schemas.push(opts.schemas[k])
            }
        } else {
            this.schemas = []
        }
    }

    /**
     *  预处理一些变量
     */
    __prepare() {
        for (let i in this.schemas) {
            let schema = this.schemas[i]
            this.schemas[i].error = ''
            if (schema.attribute.primary_key) {
                this.primaryKey = schema.column
            }
            if (schema.attribute.live.enable) {
                this._lazyFetach(schema).then(res => {
                    schema.attribute.values = res
                }).catch(e => {
                    console.log(e.message)
                })
            }
        }
    }

    _lazyFetach(schema) {
        return new Promise((resolve, reject) => {
            if (schema.attribute.live.method === 'post') {
                httpclicent.post(
                    schema.attribute.live.url,
                    schema.attribute.live.body,
                    {
                        headers: { 'Content-Type': schema.attribute.live.content_type }
                    }
                ).then(res => {
                    resolve(res)
                }).catch(e => {
                    reject(e)
                })
            } else {
                httpclicent.get(schema.attribute.live.url).then(res => {
                    resolve(res)
                }).catch(e => {
                    reject(e)
                })
            }
        })

    }

    /**
     * 刷新列表的一个模型数据
     *  @param {*} primaryKey 
     */
    __refreshModel(primaryKey) {
        return new Promise((resolve, reject) => {
            let qs = {}
            qs[this.primaryKey] = primaryKey
            qs['__format'] = 'both'
            qs['scenario'] = 'list'
            this.getModel(qs).then(res => {
                let isExists = false;
                let primaryKey = this.findModelPrimaryKey(res);
                if (primaryKey === false) {
                    reject('can not find model primary key')
                    return
                }
                for (let k in this.models) {
                    let pk = getModelValue(this.models[k], this.primaryKey);
                    if (pk == primaryKey) {
                        isExists = true
                        this.models[k] = res
                        break
                    }
                }
                if (!isExists) {
                    this.models.push(res)
                }
                resolve(res)
            }).catch(e => { reject(e) })
        })

    }

    /**
     * 删除一个模型
     * @param {*} primaryKey 
     */
    __removeModel(primaryKey) {
        for (let k in this.models) {
            let pk = getModelValue(this.models[k], this.primaryKey);
            if (primaryKey === pk) {
                this.models.splice(k, 1)
            }
        }
    }

    /**
     * 构建请求的接口地址
     * @param {String} scenario 
     * @param {any} primaryKey 
     * @returns 
     */
    __buildUri(scenario, primaryKey) {
        return this.__buildModelUri(this.modelName, this.tableName, scenario, primaryKey);
    }

    __buildModelUri(moduleName, tableNanme, scenario, primaryKey) {
        primaryKey = primaryKey || ''
        let uri = '';
        let apiPrefix = this.apiPrefix;
        let pluralizeName = pluralize.plural(tableNanme);
        let singularName = pluralize.singular(tableNanme);
        switch (scenario) {
            case 'create':
                uri += `/${apiPrefix}/${singularName}`
                break
            case 'update':
                uri += `/${apiPrefix}/${singularName}/${primaryKey}`
                break
            case 'delete':
                uri += `/${apiPrefix}/${singularName}/${primaryKey}`
                break
            case 'get':
                uri += `/${apiPrefix}/${singularName}/${primaryKey}`
                break
            case 'search':
                uri += `/${apiPrefix}/${pluralizeName}`
                break
            case 'export':
                uri += `/${apiPrefix}/${singularName}-export`
                break
            case 'import':
                uri += `/${apiPrefix}/${singularName}-import`
                break
        }
        return uri
    }

    /**
     * 初始化CRUD组件
     * @returns 
     */
    initialize() {
        return new Promise((resolve, reject) => {
            if (Array.isArray(this.schemas) && this.schemas.length > 0) {
                this.__prepare()
                resolve(this, this.schemas)
                return
            }
            if (this.modelName != "") {
                httpclicent.get(`/rest/schema/${this.modelName}/${this.tableName}`).then(res => {
                    this.schemas = res;
                    this.__prepare()
                    resolve(this, this.schemas);
                }).catch(e => {
                    reject(e)
                })
            } else {
                httpclicent.get(`/rest/schema/${this.tableName}`).then(res => {
                    this.schemas = res;
                    this.__prepare()
                    resolve(this, this.schemas);
                }).catch(e => {
                    reject(e)
                })
            }
        })
    }

    /**
     * 获取所有的字段描述信息
     * @returns 
     */
    getSchemas() {
        return this.schemas
    }

    /**
     * 设置指定列的错误信息
     * @param {string} column 
     * @param {string} error 
     */
    setColumnError(column, error) {
        for (let i in this.schemas) {
            if (this.schemas[i].column === column) {
                this.schemas[i].error = error
            }
        }
    }

    /**
     * 重置所有列的错误信息
     */
    resetError() {
        for (let i in this.schemas) {
            this.schemas[i].error = ''
        }
    }

    /**
     * 获取所有模型集合
     * @returns 
     */
    getModels() {
        return this.models;
    }

    /**
     * 设置分页索引
     * @param {Number} index 分页索引
     * @returns CRUD
     */
    setPaginationIndex(index) {
        this.pagination.index = index
        return this
    }

    /**
     * 获取当前分页的索引
     * @returns 
     */
    getPaginationIndex() {
        return this.pagination.index
    }

    /**
     * 设置分页大小
     * @returns CRUD
     * @param {Number} pageSize 分页大小
     */
    setPaginationSize(pageSize) {
        this.pagination.size = pageSize
        return this
    }

    /**
     * 获取分页大小
     * @returns 
     */
    getPaginationSize() {
        return this.pagination.size || 10
    }


    /**
     * 获取分页信息中的总数据条数
     * @returns 
     */
    getPaginationCount() {
        return this.pagination.totalCount
    }

    /**
     * 重置分页配置
     * @returns CRUD
     */
    resetPagination() {
        this.pagination.index = 1
        return this
    }

    /**
     * 设置字段排序
     * @param {String} column 字段列
     * @param {String} order  排序规则
     * @returns CRUD
     */
    setSortable(column, order) {
        if (!column) {
            this.sortable = {}
        } else {
            this.sortable = {
                column: column,
                order: order || 'ascending',
            }
        }
        return this;
    }

    /**
     * 添加固定的查询参数
     * @param {*} k 
     * @param {*} v 
     */
    addFixedQueueParams(k, v) {
        this.fixedQueueParams[k] = v;
    }

    /**
     * 设置查询参数
     * @param {Object} qs 查询参数 
     * @returns 
     */
    setQueryParams(qs) {
        this.queryParams = qs
        return this
    }

    /**
     * 添加查询参数
     * @param {String} k 
     * @param {String} v 
     */
    addQueryParams(k, v) {
        this.queryParams[k] = v;
    }

    /**
     * 查找模型的主键的值
     * @param {Object} model 表模型
     * @returns Boolean | mixed
     */
    findModelPrimaryKey(model) {
        if (typeof model === 'object') {
            return getModelValue(model, this.primaryKey);
        } else {
            return model;
        }
    }

    /**
     * 创建一个模型
     * @param {Object} model 
     */
    createModel(model) {
        return new Promise((resolve, reject) => {
            httpclicent.post(this.__buildUri('create'), model).then(res => {
                this.__refreshModel(res.id).then(res => {
                    resolve(res)
                }).catch(e => { reject(e) })
            }).catch(e => { reject(e) })
        })
    }


    /**
     * 创建一个资源
     * @param {String} moduleName 
     * @param {String} tableName 
     * @param {Object} model
     * @returns 
     */
    createResource(moduleName, tableName, model) {
        return new Promise((resolve, reject) => {
            httpclicent.post(this.__buildModelUri(moduleName, tableName, 'create'), model).then(res => {
                resolve(res)
            }).catch(e => { reject(e) })
        })
    }

    /**
     * 更新一个模型
     * @param {Object} model 
     */
    updateModel(model) {
        return new Promise((resolve, reject) => {
            let primaryKey = this.findModelPrimaryKey(model);
            if (primaryKey === false) {
                reject('can not find model primary key')
                return
            }
            httpclicent.put(this.__buildUri('update', primaryKey), model).then(res => {
                this.__refreshModel(primaryKey).then(res => {
                    resolve(res)
                }).catch(e => { reject(e) })
            }).catch(e => { reject(e) })
        })
    }

    /**
     * 更新一个资源
     * @param {String} moduleName 
     * @param {String} tableName 
     * @param {Object} model 
     */
    updateResource(moduleName, tableName, model) {
        return new Promise((resolve, reject) => {
            let primaryKey = this.findModelPrimaryKey(model);
            if (primaryKey === false) {
                reject('can not find model primary key')
                return
            }
            httpclicent.put(this.__buildModelUri(moduleName, tableName, 'update', primaryKey), model).then(res => {
                resolve(res)
            }).catch(e => { reject(e) })
        })
    }

    /**
     * 删除一个模型
     * @param {Object} model 
     */
    deleteModel(model) {
        let primaryKey = '';
        if (typeof model === 'object') {
            primaryKey = this.findModelPrimaryKey(model)
        } else {
            primaryKey = model
        }
        return new Promise((resolve, reject) => {
            httpclicent.delete(this.__buildUri('delete', primaryKey), model).then(res => {
                this.__removeModel(primaryKey)
                resolve(res)
            }).catch(e => { reject(e) })
        })
    }

    /**
     * 删除一个资源
     * @param {String} moduleName 
     * @param {String} tableName 
     * @param {Object} model 
     * @returns 
     */
    deleteResource(moduleName, tableName, model) {
        let primaryKey = '';
        if (typeof model === 'object') {
            primaryKey = this.findModelPrimaryKey(model)
        } else {
            primaryKey = model
        }
        return new Promise((resolve, reject) => {
            httpclicent.delete(this.__buildModelUri(moduleName, tableName, 'delete', primaryKey), model).then(res => {
                resolve(res)
            }).catch(e => { reject(e) })
        })
    }


    /**
     * 获取一个资源
     * @param {String} moduleName 
     * @param {String} tableName 
     * @param {Object} model 
     * @returns 
     */
    getResource(moduleName, tableName, qs) {
        return new Promise((resolve, reject) => {
            let primaryKey = '';
            let params = {}
            if (typeof qs === 'object') {
                for (let k in qs) {
                    if (k === this.primaryKey) {
                        primaryKey = qs[k]
                    } else {
                        params[k] = qs[k]
                    }
                }
            } else {
                primaryKey = qs
            }
            if (!primaryKey) {
                reject('can not find model primary key')
                return
            }
            httpclicent.get(this.__buildModelUri(moduleName, tableName, 'get', primaryKey), { params: params }).then(res => {
                resolve(res)
            }).catch(e => { reject(e) })
        })
    }

    /**
     * 
     * @param {*} pk 
     * @param {*} qs 
     * @returns 
     */
    fetchModel(pk, qs) {
        if (typeof qs !== 'object') {
            qs = {}
        }
        qs[this.primaryKey] = pk
        return this.getModel(qs)
    }

    /**
     * 获取一个模型数据
     * @param {*} qs 
     */
    getModel(qs) {
        return new Promise((resolve, reject) => {
            let primaryKey = '';
            let params = {}
            if (typeof qs === 'object') {
                for (let k in qs) {
                    if (k === this.primaryKey) {
                        primaryKey = qs[k]
                    } else {
                        params[k] = qs[k]
                    }
                }
            } else {
                primaryKey = qs
            }
            if (!primaryKey) {
                reject('can not find model primary key')
                return
            }
            httpclicent.get(this.__buildUri('get', primaryKey), { params: params }).then(res => {
                resolve(res)
            }).catch(e => { reject(e) })
        })
    }


    /**
     * 查询模型数据
     */
    searchModel() {
        let queryParams = Object.assign({}, this.queryParams);
        queryParams['page'] = this.pagination.index
        queryParams['pagesize'] = this.pagination.size || 15
        if (this.sortable && this.sortable.column != '') {
            queryParams['sort'] = this.sortable.order == 'descending' ? '-' + this.sortable.column : this.sortable.column
        }
        queryParams['__format'] = 'both';
        if (typeof this.fixedQueueParams === 'object') {
            for (let k in this.fixedQueueParams) {
                queryParams[k] = this.fixedQueueParams[k];
            }
        }
        return new Promise((resolve, reject) => {
            httpclicent.get(this.__buildUri('search'), { params: queryParams }).then(res => {
                this.pagination.index = parseInt(res.page);
                this.pagination.size = parseInt(res.pagesize);
                this.pagination.totalCount = parseInt(res.totalCount);
                this.models = res.data
                resolve()
            }).catch(e => { reject(e) })
        })
    }

    /**
     * 批量删除模型
     * @param {Array} data 
     */
    deleteModels(data) {
        let pks = [];
        for (let val of data) {
            if (typeof val === 'object') {
                let pk = this.findModelPrimaryKey(val)
                if (pk) {
                    pks.push(pk)
                }
            } else {
                pks.push(val)
            }
        }
        if (pks.length <= 0) {
            return
        }
        let ps = [];
        for (let pk of pks) {
            ps.push(this.deleteModel(pk))
        }
        return new Promise((resolve, reject) => {
            Promise.all(ps).then(res => {
                let data = {
                    total: 0,
                    success: 0,
                    responses: res,
                };
                data.total = pks.length
                data.success = res.length
                resolve(data)
            }).catch(e => {
                reject(e)
            })
        })
    }

    /**
     * 下载指定资源的导入模板
     * @param {String} moduleName 
     * @param {String} tableName 
     * @returns 
     */
    downloadResourceTemplate(moduleName, tableName) {
        return new Promise((resolve, reject) => {
            httpclicent.get(this.__buildModelUri(moduleName, tableName, 'import')).then(res => {
                let element = document.createElement('a')
                let disposition = res.headers['content-disposition']
                let filename = this.tableName + '.csv';
                //获取文件名称
                if (disposition) {
                    let pos = -1;
                    let ps = disposition.split(';')
                    for (let s of ps) {
                        pos = s.indexOf('filename=')
                        if (pos === 0) {
                            filename = s.substring(9)
                            break
                        }
                    }
                }
                element.style.display = 'none'
                element.href = window.URL.createObjectURL(new Blob([res.data], { type: res.headers['content-type'] }))
                element.target = '_blank'
                element.setAttribute('download', filename)
                document.body.appendChild(element)
                element.click()
                window.URL.revokeObjectURL(element.href)
                document.body.removeChild(element)
            }).catch(e => { reject(e) })
        })
    }

    /**
     * 下载导入模板
     * @returns 
     */
    downloadTemplate() {
        return this.downloadResourceTemplate(this.modelName, this.tableName);
    }

    /**
     * 导出模型数据
     * @returns 
     */
    exportModels() {
        let queryParams = Object.assign({}, this.queryParams);
        if (this.sortable && this.sortable.column != '') {
            queryParams['sort'] = this.sortable.order == 'descending' ? '-' + this.sortable.column : this.sortable.column
        }
        return new Promise((resolve, reject) => {
            httpclicent.get(this.__buildUri('export'), { params: queryParams }).then(res => {
                let element = document.createElement('a')
                let disposition = res.headers['content-disposition']
                let filename = this.tableName + '.csv';
                //获取文件名称
                if (disposition) {
                    let pos = -1;
                    let ps = disposition.split(';')
                    for (let s of ps) {
                        pos = s.indexOf('filename=')
                        if (pos === 0) {
                            filename = s.substring(9)
                            break
                        }
                    }
                }
                element.style.display = 'none'
                element.href = window.URL.createObjectURL(new Blob([res.data], { type: res.headers['content-type'] }))
                element.target = '_blank'
                element.setAttribute('download', filename)
                document.body.appendChild(element)
                element.click()
                window.URL.revokeObjectURL(element.href)
                document.body.removeChild(element)
            }).catch(e => { reject(e) })
        })
    }


}

export default CRUD